# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)


import re
import logging
import threading
from PySide2 import QtCore
import subprocess

import config
import user_login
import version


class CoreModel(QtCore.QObject):

    PYTHON_QML_CONSTANTS_FILE = "qt_interface/qml/constants.js"

    stateChanged = QtCore.Signal()
    on_show_hello = QtCore.Signal()
    on_offlineModeChanged = QtCore.Signal()
    on_cameraState = QtCore.Signal()

    showMessage = QtCore.Signal(str, str, str)

    on_updateCheck = QtCore.Signal(str)
    on_updateError = QtCore.Signal(str)

    def __init__(self, app):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self._app = app
        self._cameraEnabled = False
        self._loadConstants()
        self._state = self.STATE_CONNECTING
        self._previous_state = self.STATE_CONNECTING
        self._state_lock = threading.Lock()
        self._show_hello = config.get_settings()['wizard_on_start']

    def _constantNameToClassPrefix(self, name):
        new_name = ""
        previous_letter = ""
        for letter in name:
            if letter.isupper() and previous_letter.islower():
                new_name += "_"
            new_name += letter
            previous_letter = letter
        new_name = new_name.upper()
        new_name = new_name.rstrip("S")
        new_name = new_name + "_" 
        return new_name



    def _loadConstants(self):
        with open(self.PYTHON_QML_CONSTANTS_FILE) as f:
            current_var = None
            text = f.read()
            for line in text.split("\n"):
                var_search = re.match("\s*var(.*) = ", line)
                if not line:
                    continue
                if var_search:
                    current_var = var_search.group(1).strip()
                if current_var:
                    if "}" in line:
                        current_var = None
                    elif ":" in line:
                        line = line.replace(",", "")
                        attr_name, attr_value = line.split(":")
                        prefix = self._constantNameToClassPrefix(current_var)
                        attr_name = prefix + attr_name.strip()
                        self.logger.debug("QML constant %s set to attribute %s" % (attr_value, attr_name))
                        setattr(self, attr_name, attr_value.strip().strip('"'))

    def setState(self, state):
        with self._state_lock:
            self._previous_state = self._state
            self._state = state
            if self._previous_state != state:
                self.logger.debug("Model state changed from %s to %s " % (self._previous_state, state))
                self.stateChanged.emit()

    def _checkForUpdates(self):
        updater = self._app.updater
        updater.check_for_updates()
        self.on_updateCheck.emit(updater.update_download_url)

    def _updateCameraState(self):
        cameraName = self._app.camera_controller.get_current_camera_name()
        self._cameraEnabled = cameraName != "Disable camera"

    @QtCore.Slot()
    def shutdown(self):
        self.logger.warning("Shutdown")
        self._app.poweroff_flag = True
        self._app.stop_flag = True

    @QtCore.Slot()
    def unbind(self):
        self.logger.warning("Unbinding account(erasing token)")
        user_login.UserLogin.logout()
        self._app.user_login.forget_auth_tokens()
        self._app.qt_interface.printer_model.disconnect_printer()

    @QtCore.Slot()
    def toggleOfflineMode(self):
        self.logger.info("Offiline mode toggled")
        offline_mode = not self._app.offline_mode
        self._app.set_offline_mode(offline_mode)
        self.on_offlineModeChanged.emit()

    @QtCore.Slot()
    def switchCloud(self):
        settings = config.get_settings()
        url = settings["URL"].replace("cli-", "")
        if url == "cloud.3dprinteros.com":
            config.Config.instance().update_settings({"URL": "acorn.3dprinteros.com"})
        else:
            config.Config.instance().update_settings({"URL": "cli-cloud.3dprinteros.com"})
        self.shutdown()

    @QtCore.Slot()
    def toggleSSH(self):
        if not self._app.is_sudo_available():
            self.logger.info("No passwordless sudo to toggle sshd(checked by username).")
        else:
            try:
                output = subprocess.check_output(["systemctl", "is-active", "sshd"])
            except subprocess.CalledProcessError:
                output = b"inactive" # exit code 3 means that service in inactive
            try:
                if b"inactive" in output:
                    self.logger.info("Starting and enabling ssh...")
                    subprocess.call(["sudo", "systemctl", "start", "sshd"])
                    subprocess.call(["sudo", "systemctl", "enable", "sshd"])
                    self.logger.info("ssh enabled")
                else:
                    self.logger.info("Stopping and disabling ssh...")
                    subprocess.call(["sudo", "systemctl", "stop", "sshd"])
                    subprocess.call(["sudo", "systemctl", "disable", "sshd"])
                    self.logger.info("ssh disabled")
            except (subprocess.CalledProcessError, OSError):
                self.logger.exception("Exception while toggling sshd")

    @QtCore.Slot()
    def checkForUpdates(self):
        thread = threading.Thread(target=self._checkForUpdates)
        thread.start()

    @QtCore.Slot()
    def performUpdate(self):
        error = self._app.updater.download_update()
        if error:
            self.on_updateError(error)
        else:
            self._app.stop_flag = True

    @QtCore.Slot()
    def enableCamera(self, newCameraTypeName="Pi camera"):
        configObject = config.Config.instance()
        cameraSettingsDict = config.get_settings()["camera"]
        cameraSettingsDict["default"] = newCameraTypeName
        configObject.update_settings({"camera": cameraSettingsDict})
        self._app.camera_controller.start_camera_process(newCameraTypeName)
        self._updateCameraState()
        self.on_cameraState.emit()

    @QtCore.Slot()
    def disableCamera(self):
        configObject = config.Config.instance()
        cameraSettingsDict = config.get_settings()["camera"]
        cameraSettingsDict["default"] = "Disable camera"
        configObject.update_settings({"camera": cameraSettingsDict})
        self._app.camera_controller.stop_camera_process()
        self._updateCameraState()
        self.on_cameraState.emit()

    @QtCore.Slot()
    def toggleVirtualPrinter(self):
        configObject = config.Config.instance()
        if self._app.virtual_printer_enabled:
            self._app.virtual_printer_enabled = False
            configObject.update_settings({"virtual_printer": { "enabled": False }})
        else:
            self._app.virtual_printer_enabled = True
            configObject.update_settings({"virtual_printer": { "enabled": True }})

    # @QtCore.Slot()
    # def touchCalibration(self):
        # #OPT --precalib min_x=0, max_x=65535 and min_y=0, max_y=65535
        # process = subprocess.run(['xinput_calibrator', '--output-type', 'xorg.conf.d'], capture_output=True)
        # calibration_config_lines = []
        # for line in process.stdout.split(b"\n"):
            # if calibration_config_lines:
                # calibration_config_lines.append(line)
            # elif b"EndSection" in line:
                # break
            # elif b"Section" in line:
                # calibration_config_lines.append(line)
                # calibration_config_lines.append(b'Driver "evdev"')
        # calibration_config = b"\n".join(calibration_config_lines)
        # output_filename = '/tmp/99-calibration.conf'
        # with open(output_filename, 'wb') as f:
            # f.write(calibration_config)
        # if config.get_settings()['subprocesses']['sudo']:
            # subprocess.run(['sudo', 'cp', output_filename, '/etc/X11/xorg.conf.d/'])
            # subprocess.run(['sync'])

    @QtCore.Property(str, notify=stateChanged)
    def state(self):
        return self._state

    @QtCore.Property(bool, notify=on_show_hello)
    def show_hello(self):
        return self._show_hello

    @QtCore.Property(bool, notify=on_offlineModeChanged)
    def offlineMode(self):
        return self._app.offline_mode

    @QtCore.Property(str, constant=True)
    def version(self):
        return version.version + " " + version.build

    @QtCore.Property(str, constant=True)
    def cloudURL(self):
        return config.get_settings()["URL"].replace("cli-", "")

    @QtCore.Property(bool, notify=on_cameraState)
    def cameraState(self):
        self._updateCameraState()
        return self._cameraEnabled
